/*
 * Interrupt handling.
 *
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef CONFIG_CPU_JZ4730_MINIPC
#include "minipc.h"
#else
#include "nanonote.h"
#endif

#include "board-specific.h"
#include "board.h"
#include "lcd.h"
#include "jzlcd.h"
#include "cpu.h"
#include "irq.h"

extern vidinfo_t panel_info;

void next_pixel(unsigned short *x, unsigned short *y)
{
	(*x)++;
	if (*x >= panel_info.vl_col) {
		*x = 0;
		(*y)++;
		if (*y >= panel_info.vl_row)
			*y = 0;
	}
}

/* Task management. */

enum { max_tasks = 3 };
static u32 stack_pointers[max_tasks];
static u32 registers[max_tasks][32];

u8 current_task;
u32 *current_stack_pointer;
u32 *current_registers;

extern u32 _got_copy_start;

const u32 stack_start = 0x00080000;
const u32 stack_size = 0x00002000;
const u32 pagesize = 4 * 1024;

/* Tasks. */

void plot_pattern(unsigned short pixel_type, unsigned short x, unsigned short y)
{
	while (1) {
		if (pixel_type)
			test_pixel(x, y, pixel_type);
		else
			clear_pixel(x, y);
		next_pixel(&x, &y);
		udelay(100);
	}
}

/* Initialisation and handling. */

void irq_init()
{
	handle_error_level();
	timer_init_irq();
	init_interrupts();
}

void irq_handle()
{
	unsigned short i;

	/* Check interrupt identity. */

	if (REG_INTC_IPR & (1 << TIMER_CHAN_IRQ)) {

		/* Switch task. */

		switch_task();

		/* Clear interrupt status. */

		__intc_ack_irq(TIMER_CHAN_IRQ);
		__tcu_clear_full_match_flag(TIMER_CHAN);

	/* Handle other interrupts, anyway. */

	} else {
		for (i = 0; i < 32; i++) {
			if (REG_INTC_IPR & (1 << i))
				__intc_ack_irq(i);
		}
	}
}

void init_tasks()
{
	current_task = 0;
	current_stack_pointer = &stack_pointers[current_task];
	current_registers = registers[current_task];
}

void start_task(unsigned short task)
{
	u32 args[] = {task, 0, (task - 1) * 60};
	u32 virtual, physical;

	/*
	Each task employs a stack at a multiple of the given start address in
	physical memory, but at the same address in virtual memory.
	*/

	physical = stack_start - stack_size * task;
	virtual = physical;

	init_page_table(page_table_start, virtual - pagesize * 2, physical - pagesize * 2, pagesize, 0x1e, task);

	stack_pointers[task] = virtual - 12;

	/*
	Set the registers for the new task, initialising the global pointer and
	return address.
	*/

	init_registers(registers[task], _got_copy_start, (void (*)()) plot_pattern, args, 3);
}

void switch_task()
{
	/* Switch the current task. */

	current_task++;
	if (current_task == max_tasks) current_task = 1;

	/* Indicate the current stack pointer and task registers. */

	current_stack_pointer = &stack_pointers[current_task];
	current_registers = registers[current_task];
}
